@@ -64,9 +64,7 @@ static const struct got_error *gotweb_get_clone_url(char **, struct server *,
 
 static void gotweb_free_repo_dir(struct repo_dir *);
 
-struct server *gotweb_get_server(const char *);
-
-static int
+int
 gotweb_reply(struct request *c, int status, const char *ctype,
     struct gotweb_url *location)
 {
@@ -77,7 +75,7 @@ gotweb_reply(struct request *c, int status, const char *ctype,
 
 	if (location) {
 		if (tp_writes(c->tp, "Location: ") == -1 ||
-		    gotweb_render_url(c, location) == -1 ||
+		    gotweb_render_absolute_url(c, location) == -1 ||
 		    tp_writes(c->tp, "\r\n") == -1)
 			return -1;
 	}
@@ -126,7 +124,7 @@ cleanup_request(struct request *c)
 
 	fcgi_cleanup_request(c);
 
-	if (imsg_compose_event(gotwebd_env->iev_sockets, GOTWEBD_IMSG_REQ_ABORT,
+	if (imsg_compose_event(gotwebd_env->iev_auth, GOTWEBD_IMSG_REQ_ABORT,
 	    GOTWEBD_PROC_GOTWEB, -1, -1, &request_id, sizeof(request_id)) == -1)
 		log_warn("imsg_compose_event");
 }
@@ -539,6 +537,28 @@ gotweb_init_transport(struct transport **t)
 	return error;
 }
 
+struct gotwebd_repo *
+gotweb_get_repository(struct server *server, const char *name)
+{
+	struct gotwebd_repo *repo;
+
+	TAILQ_FOREACH(repo, &server->repos, entry) {
+		if (strncmp(repo->name, name, strlen(repo->name)) != 0)
+			continue;
+	
+		if (strlen(name) == strlen(repo->name))
+			return repo;
+
+		if (strlen(name) != strlen(repo->name) + 4)
+			continue;
+
+		if (strcmp(name + strlen(repo->name), ".git") == 0)
+			return repo;
+	}
+
+	return NULL;
+}
+
 void
 gotweb_free_repo_tag(struct repo_tag *rt)
 {
@@ -958,16 +978,49 @@ validate_path(const char *path, const char *parent_path,
 	return error;
 }
 
+static enum gotwebd_access
+auth_check(struct request *c, struct gotwebd_access_rule_list *rules)
+{
+	struct gotwebd *env = gotwebd_env;
+	enum gotwebd_access access = GOTWEBD_ACCESS_NO_MATCH;
+	struct gotwebd_access_rule *rule;
+
+	/*
+	 * The www user ID implies that no user authentication occurred.
+	 * But authentication is enabled so we must deny this request.
+	 */
+	if (c->client_uid == env->www_uid)
+		return GOTWEBD_ACCESS_DENIED;
+
+	/*
+	 * We cannot access /etc/passwd in this process so we cannot
+	 * verify the client's user ID outselves here.
+	 * Match rules against the access identifier which has already
+	 * passed authentication in the auth process.
+	 */
+	STAILQ_FOREACH(rule, rules, entry) {
+		if (strcmp(rule->identifier, c->access_identifier) == 0)
+			access = rule->access;
+	}
+
+	return access;
+}
+
 static const struct got_error *
 gotweb_load_got_path(struct repo_dir **rp, const char *dir,
     struct request *c)
 {
 	const struct got_error *error = NULL;
+	struct gotwebd *env = gotwebd_env;
 	struct server *srv = c->srv;
 	struct transport *t = c->t;
 	struct repo_dir *repo_dir;
 	DIR *dt;
 	char *dir_test;
+	struct gotwebd_repo *repo;
+	enum gotwebd_auth_config auth_config = 0;
+	enum gotwebd_access access = GOTWEBD_ACCESS_DENIED;
+	int repo_is_hidden = 0;
 
 	*rp = calloc(1, sizeof(**rp));
 	if (*rp == NULL)
@@ -1007,9 +1060,58 @@ gotweb_load_got_path(struct repo_dir **rp, const char *dir,
 	repo_dir->path = dir_test;
 	dir_test = NULL;
 
-	repo_dir->name = strdup(repo_dir->path + strlen(srv->repos_path) + 1);
-	if (repo_dir->name == NULL) {
-		error = got_error_from_errno("strdup");
+	error = got_path_basename(&repo_dir->name, repo_dir->path);
+	if (error)
+		goto err;
+
+	repo = gotweb_get_repository(srv, repo_dir->name);
+	if (repo) {
+		repo_is_hidden = repo->hidden;
+		auth_config = repo->auth_config;
+		switch (auth_config) {
+		case GOTWEBD_AUTH_DISABLED:
+			access = GOTWEBD_ACCESS_PERMITTED;
+			break;
+		case GOTWEBD_AUTH_SECURE:
+		case GOTWEBD_AUTH_INSECURE:
+			access = auth_check(c, &repo->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &srv->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &env->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = GOTWEBD_ACCESS_DENIED;
+			break;
+		}
+	} else {
+		repo_is_hidden = srv->hide_repositories;
+		auth_config = srv->auth_config;
+		switch (auth_config) {
+		case GOTWEBD_AUTH_DISABLED:
+			access = GOTWEBD_ACCESS_PERMITTED;
+			break;
+		case GOTWEBD_AUTH_SECURE:
+		case GOTWEBD_AUTH_INSECURE:
+			access = auth_check(c, &srv->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = auth_check(c, &env->access_rules);
+			if (access == GOTWEBD_ACCESS_NO_MATCH)
+				access = GOTWEBD_ACCESS_DENIED;
+			break;
+		}
+	}
+
+	if (access != GOTWEBD_ACCESS_PERMITTED &&
+	    access != GOTWEBD_ACCESS_DENIED)
+		fatalx("invalid access check result %d", access);
+
+	if (access != GOTWEBD_ACCESS_PERMITTED) {
+		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
+		goto err;
+	}
+
+	if (repo_is_hidden) {
+		error = got_error_path(repo_dir->name, GOT_ERR_NOT_GIT_REPO);
 		goto err;
 	}
 
@@ -1171,14 +1273,19 @@ gotweb_shutdown(void)
 	imsgbuf_clear(&gotwebd_env->iev_parent->ibuf);
 	free(gotwebd_env->iev_parent);
 
-	imsgbuf_clear(&gotwebd_env->iev_sockets->ibuf);
-	free(gotwebd_env->iev_sockets);
+	imsgbuf_clear(&gotwebd_env->iev_auth->ibuf);
+	free(gotwebd_env->iev_auth);
+
+	config_free_access_rules(&gotwebd_env->access_rules);
 
 	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
 		struct server *srv;
 
 		srv = TAILQ_FIRST(&gotwebd_env->servers);
 		TAILQ_REMOVE(&gotwebd_env->servers, srv, entry);
+
+		config_free_access_rules(&srv->access_rules);
+		config_free_repos(&srv->repos);
 		free(srv);
 	}
 
@@ -1190,6 +1297,14 @@ gotweb_shutdown(void)
 		free(sock);
 	}
 
+	while (!TAILQ_EMPTY(&gotwebd_env->addresses)) {
+		struct address *h;
+
+		h = TAILQ_FIRST(&gotwebd_env->addresses);
+		TAILQ_REMOVE(&gotwebd_env->addresses, h, entry);
+		free(h);
+	}
+
 	free(gotwebd_env);
 
 	exit(0);
@@ -1225,8 +1340,8 @@ gotweb_launch(struct gotwebd *env)
 	struct server *srv;
 	const struct got_error *error;
 
-	if (env->iev_sockets == NULL)
-		fatal("sockets process not connected");
+	if (env->iev_auth == NULL)
+		fatal("auth process not connected");
 
 #ifndef PROFILE
 	if (pledge("stdio rpath recvfd sendfd proc exec unveil", NULL) == -1)
@@ -1245,7 +1360,7 @@ gotweb_launch(struct gotwebd *env)
 	if (unveil(NULL, NULL) == -1)
 		fatal("unveil");
 
-	event_add(&env->iev_sockets->ev, NULL);
+	event_add(&env->iev_auth->ev, NULL);
 }
 
 static void
@@ -1314,14 +1429,17 @@ gotweb_dispatch_server(int fd, short event, void *arg)
 }
 
 static void
-recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
+recv_auth_pipe(struct gotwebd *env, struct imsg *imsg)
 {
 	struct imsgev *iev;
 	int fd;
 
+	if (env->iev_auth != NULL)
+		fatalx("auth process already connected");
+
 	fd = imsg_get_fd(imsg);
 	if (fd == -1)
-		fatalx("invalid server pipe fd");
+		fatalx("invalid auth pipe fd");
 
 	iev = calloc(1, sizeof(*iev));
 	if (iev == NULL)
@@ -1335,7 +1453,7 @@ recv_server_pipe(struct gotwebd *env, struct imsg *imsg)
 	event_set(&iev->ev, fd, EV_READ, gotweb_dispatch_server, iev);
 	imsg_event_add(iev);
 
-	env->iev_sockets = iev;
+	env->iev_auth = iev;
 }
 
 static void
@@ -1345,6 +1463,8 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 	struct imsgbuf		*ibuf;
 	struct imsg		 imsg;
 	struct gotwebd		*env = gotwebd_env;
+	struct server		*srv;
+	struct gotwebd_repo	*repo;
 	ssize_t			 n;
 	int			 shut = 0;
 
@@ -1368,9 +1488,35 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 			break;
 
 		switch (imsg.hdr.type) {
+		case GOTWEBD_IMSG_CFG_ACCESS_RULE:
+			if (TAILQ_EMPTY(&env->servers)) {
+				/* global access rule */
+				config_get_access_rule(&env->access_rules,
+				    &imsg);
+			} else {
+				srv = TAILQ_LAST(&env->servers, serverlist);
+				if (TAILQ_EMPTY(&srv->repos)) {
+					/* per-server access rule */
+					config_get_access_rule(
+					    &srv->access_rules, &imsg);
+				} else {
+					/* per-repository access rule */
+					repo = TAILQ_LAST(&srv->repos,
+					    gotwebd_repolist);
+					config_get_access_rule(
+					    &repo->access_rules, &imsg);
+				}
+			}
+			break;
 		case GOTWEBD_IMSG_CFG_SRV:
 			config_getserver(env, &imsg);
 			break;
+		case GOTWEBD_IMSG_CFG_REPO:
+			if (TAILQ_EMPTY(&env->servers))
+				fatalx("%s: unexpected CFG_REPO msg", __func__);
+			srv = TAILQ_LAST(&env->servers, serverlist);
+			config_get_repository(&srv->repos, &imsg);
+			break;
 		case GOTWEBD_IMSG_CFG_FD:
 			config_getfd(env, &imsg);
 			break;
@@ -1381,7 +1527,17 @@ gotweb_dispatch_main(int fd, short event, void *arg)
 			config_getcfg(env, &imsg);
 			break;
 		case GOTWEBD_IMSG_CTL_PIPE:
-			recv_server_pipe(env, &imsg);
+			recv_auth_pipe(env, &imsg);
+			break;
+		case GOTWEBD_IMSG_AUTH_CONF:
+			if (imsg_get_data(&imsg, &env->auth_config,
+			    sizeof(env->auth_config)) == -1)
+				fatalx("%s: invalid AUTH_CONF msg", __func__);
+			break;
+		case GOTWEBD_IMSG_WWW_UID:
+			if (imsg_get_data(&imsg, &env->www_uid,
+			    sizeof(env->www_uid)) == -1)
+				fatalx("%s: invalid WWW_UID msg", __func__);
 			break;
 		case GOTWEBD_IMSG_CTL_START:
 			gotweb_launch(env);
