@@ -55,6 +55,8 @@ void	 gotwebd_sighdlr(int sig, short event, void *arg);
 void	 gotwebd_shutdown(void);
 void	 gotwebd_dispatch_server(int, short, void *);
 void	 gotwebd_dispatch_fcgi(int, short, void *);
+void	 gotwebd_dispatch_login(int, short, void *);
+void	 gotwebd_dispatch_auth(int, short, void *);
 void	 gotwebd_dispatch_gotweb(int, short, void *);
 
 struct gotwebd	*gotwebd_env;
@@ -124,6 +126,13 @@ main_compose_sockets(struct gotwebd *env, uint32_t type, int fd,
 	return send_imsg(env->iev_sockets, type, fd, data, len);
 }
 
+int
+main_compose_login(struct gotwebd *env, uint32_t type, int fd,
+    const void *data, uint16_t len)
+{
+	return send_imsg(env->iev_login, type, fd, data, len);
+}
+
 int
 main_compose_gotweb(struct gotwebd *env, uint32_t type, int fd,
     const void *data, uint16_t len)
@@ -140,6 +149,22 @@ main_compose_gotweb(struct gotwebd *env, uint32_t type, int fd,
 	return ret;
 }
 
+int
+main_compose_auth(struct gotwebd *env, uint32_t type, int fd,
+    const void *data, uint16_t len)
+{
+	size_t i;
+	int ret = 0;
+
+	for (i = 0; i < env->prefork; i++) {
+		ret = send_imsg(&env->iev_auth[i], type, fd, data, len);
+		if (ret)
+			break;
+	}
+
+	return ret;
+}
+
 int
 sockets_compose_main(struct gotwebd *env, uint32_t type, const void *d,
     uint16_t len)
@@ -147,6 +172,52 @@ sockets_compose_main(struct gotwebd *env, uint32_t type, const void *d,
 	return (imsg_compose_event(env->iev_parent, type, 0, -1, -1, d, len));
 }
 
+ void
+gotwebd_dispatch_login(int fd, short event, void *arg)
+{
+	struct imsgev		*iev = arg;
+	struct imsgbuf		*ibuf;
+	struct imsg		 imsg;
+	ssize_t			 n;
+	int			 shut = 0;
+
+	ibuf = &iev->ibuf;
+
+	if (event & EV_READ) {
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
+		if (n == 0)	/* Connection closed */
+			shut = 1;
+	}
+	if (event & EV_WRITE) {
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
+	}
+
+	for (;;) {
+		if ((n = imsg_get(ibuf, &imsg)) == -1)
+			fatal("imsg_get");
+		if (n == 0)	/* No more messages. */
+			break;
+
+		switch (imsg.hdr.type) {
+		default:
+			fatalx("%s: unknown imsg type %d", __func__,
+			    imsg.hdr.type);
+		}
+
+		imsg_free(&imsg);
+	}
+
+	if (!shut)
+		imsg_event_add(iev);
+	else {
+		/* This pipe is dead.  Remove its event handler */
+		event_del(&iev->ev);
+		event_loopexit(NULL);
+	}
+}
+
 void
 gotwebd_dispatch_server(int fd, short event, void *arg)
 {
@@ -247,6 +318,56 @@ gotwebd_dispatch_fcgi(int fd, short event, void *arg)
 	}
 }
 
+void
+gotwebd_dispatch_auth(int fd, short event, void *arg)
+{
+	struct imsgev		*iev = arg;
+	struct imsgbuf		*ibuf;
+	struct imsg		 imsg;
+	struct gotwebd		*env = gotwebd_env;
+	ssize_t			 n;
+	int			 shut = 0;
+
+	ibuf = &iev->ibuf;
+
+	if (event & EV_READ) {
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
+		if (n == 0)	/* Connection closed */
+			shut = 1;
+	}
+	if (event & EV_WRITE) {
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
+	}
+
+	for (;;) {
+		if ((n = imsg_get(ibuf, &imsg)) == -1)
+			fatal("imsg_get");
+		if (n == 0)	/* No more messages. */
+			break;
+
+		switch (imsg.hdr.type) {
+		case GOTWEBD_IMSG_CFG_DONE:
+			gotwebd_configure_done(env);
+			break;
+		default:
+			fatalx("%s: unknown imsg type %d", __func__,
+			    imsg.hdr.type);
+		}
+
+		imsg_free(&imsg);
+	}
+
+	if (!shut)
+		imsg_event_add(iev);
+	else {
+		/* This pipe is dead.  Remove its event handler */
+		event_del(&iev->ev);
+		event_loopexit(NULL);
+	}
+}
+
 void
 gotwebd_dispatch_gotweb(int fd, short event, void *arg)
 {
@@ -324,14 +445,29 @@ gotwebd_sighdlr(int sig, short event, void *arg)
 static void
 spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
     enum gotwebd_proc_type proc_type, const char *username,
-    void (*handler)(int, short, void *))
+    const char *www_user, void (*handler)(int, short, void *))
 {
 	const char	*argv[10];
 	int		 argc = 0;
 	int		 p[2];
 	pid_t		 pid;
+	char		 usernames[_PW_NAME_LEN * 2 + 1 + 1];
+	int		 ret;
+	int		 sock_flags = SOCK_STREAM | SOCK_NONBLOCK;
 
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, p) == -1)
+#ifdef SOCK_CLOEXEC
+	sock_flags |= SOCK_CLOEXEC;
+#endif
+
+	ret = snprintf(usernames, sizeof(usernames), "%s:%s",
+	    username, www_user);
+	if (ret == -1)
+		fatal("snprintf");
+	if ((size_t)ret >= sizeof(usernames))
+		fatalx("usernames too long");
+
+	if (socketpair(AF_UNIX,
+	    sock_flags, PF_UNSPEC, p) == -1)
 		fatal("socketpair");
 
 	switch (pid = fork()) {
@@ -358,16 +494,22 @@ spawn_process(struct gotwebd *env, const char *argv0, struct imsgev *iev,
 		char *s;
 
 		argv[argc++] = "-S";
-		argv[argc++] = username;
+		argv[argc++] = usernames;
 		if (asprintf(&s, "-S%d", env->prefork) == -1)
 			fatal("asprintf");
 		argv[argc++] = s;
+	} else if (proc_type == GOTWEBD_PROC_LOGIN) {
+		argv[argc++] = "-L";
+		argv[argc++] = usernames;
 	} else if (proc_type == GOTWEBD_PROC_FCGI) {
 		argv[argc++] = "-F";
-		argv[argc++] = username;
+		argv[argc++] = usernames;
+	} else if (proc_type == GOTWEBD_PROC_AUTH) {
+		argv[argc++] = "-A";
+		argv[argc++] = usernames;
 	} else if (proc_type == GOTWEBD_PROC_GOTWEB) {
 		argv[argc++] = "-G";
-		argv[argc++] = username;
+		argv[argc++] = usernames;
 	}
 	if (strcmp(env->gotwebd_conffile, GOTWEBD_CONF) != 0) {
 		argv[argc++] = "-f";
@@ -400,6 +542,26 @@ usage(void)
 	exit(1);
 }
 
+static void
+get_usernames(const char **gotwebd_username, const char **www_username,
+    char *optarg)
+{
+	static char usernames[_PW_NAME_LEN * 2 + 1 + 1];
+	char *colon;
+
+	if (strlcpy(usernames, optarg, sizeof(usernames)) >=
+	    sizeof(usernames))
+		fatalx("usernames too long");
+
+	colon = strchr(usernames, ':');
+	if (colon == NULL)
+		fatalx("bad username option parameter");
+	*colon = '\0';
+
+	*gotwebd_username = &usernames[0];
+	*www_username = colon + 1;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -428,8 +590,12 @@ main(int argc, char **argv)
 		fatal("%s: calloc", __func__);
 	config_init(env);
 
-	while ((ch = getopt(argc, argv, "D:dG:f:F:nS:vW:")) != -1) {
+	while ((ch = getopt(argc, argv, "A:D:dG:f:F:L:nS:vW:")) != -1) {
 		switch (ch) {
+		case 'A':
+			proc_type = GOTWEBD_PROC_AUTH;
+			get_usernames(&gotwebd_username, &www_username, optarg);
+			break;
 		case 'D':
 			if (cmdline_symset(optarg) < 0)
 				log_warnx("could not parse macro definition %s",
@@ -440,14 +606,18 @@ main(int argc, char **argv)
 			break;
 		case 'G':
 			proc_type = GOTWEBD_PROC_GOTWEB;
-			gotwebd_username = optarg;
+			get_usernames(&gotwebd_username, &www_username, optarg);
 			break;
 		case 'f':
 			conffile = optarg;
 			break;
 		case 'F':
 			proc_type = GOTWEBD_PROC_FCGI;
-			gotwebd_username = optarg;
+			get_usernames(&gotwebd_username, &www_username, optarg);
+			break;
+		case 'L':
+			proc_type = GOTWEBD_PROC_LOGIN;
+			get_usernames(&gotwebd_username, &www_username, optarg);
 			break;
 		case 'n':
 			no_action = 1;
@@ -455,9 +625,10 @@ main(int argc, char **argv)
 		case 'S':
 			proc_type = GOTWEBD_PROC_SOCKETS;
 			i = strtonum(optarg, 1, INT_MAX, &errstr);
-			if (errstr)
-				gotwebd_username = optarg;
-			else
+			if (errstr) {
+				get_usernames(&gotwebd_username,
+				    &www_username, optarg);
+			} else
 				env->prefork = i;
 			break;
 		case 'v':
@@ -501,23 +672,44 @@ main(int argc, char **argv)
 	pw = getpwnam(www_username);
 	if (pw == NULL)
 		fatalx("unknown user %s", www_username);
+	env->www_uid = pw->pw_uid;
 	www_gid = pw->pw_gid;
+	if (pw->pw_uid == 0 || pw->pw_gid == 0) {
+		warnx("warning: detected www user \"root\" in gotwebd.conf; "
+		    "running a web server with UID/GID 0 is dangerous");
+	}
 
 	pw = getpwnam(gotwebd_username);
 	if (pw == NULL)
 		fatalx("unknown user %s", gotwebd_username);
+	if (pw->pw_uid == 0 || pw->pw_gid == 0) {
+		fatalx("refusing to start up with user \"root\" set "
+		    "in gotwebd.conf; running gotwebd with UID/GID 0 "
+		    "is dangerous");
+	}
 	if (getgrouplist(gotwebd_username, pw->pw_gid, gotwebd_groups,
 	    &gotwebd_ngroups) == -1)
 		fatalx("too many groups for user %s", gotwebd_username);
 
 	/* check for root privileges */
 	if (geteuid())
-		fatalx("need root privileges");
+		fatalx("need root privileges to start up");
 
 	log_init(env->gotwebd_debug, LOG_DAEMON);
 	log_setverbose(env->gotwebd_verbose);
 
 	switch (proc_type) {
+	case GOTWEBD_PROC_LOGIN:
+		setproctitle("login");
+		log_procinit("login");
+
+		if (setgroups(1, &pw->pw_gid) == -1 ||
+		    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) == -1 ||
+		    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid) == -1)
+			fatal("failed to drop privileges");
+
+		gotwebd_login(env, GOTWEBD_SOCK_FILENO);
+		return 1;
 	case GOTWEBD_PROC_SOCKETS:
 		setproctitle("sockets");
 		log_procinit("sockets");
@@ -550,6 +742,17 @@ main(int argc, char **argv)
 
 		gotwebd_fcgi(env, GOTWEBD_SOCK_FILENO);
 		return 1;
+	case GOTWEBD_PROC_AUTH:
+		setproctitle("auth");
+		log_procinit("auth");
+
+		if (setgroups(1, &pw->pw_gid) == -1 ||
+		    setresgid(pw->pw_gid, pw->pw_gid, pw->pw_gid) == -1 ||
+		    setresuid(pw->pw_uid, pw->pw_uid, pw->pw_uid) == -1)
+			fatal("failed to drop privileges");
+
+		gotwebd_auth(env, GOTWEBD_SOCK_FILENO);
+		return 1;
 	case GOTWEBD_PROC_GOTWEB:
 		setproctitle("gotweb");
 		log_procinit("gotweb");
@@ -574,25 +777,39 @@ main(int argc, char **argv)
 	if (env->iev_sockets == NULL)
 		fatal("calloc");
 
+	env->iev_login = calloc(1, sizeof(*env->iev_login));
+	if (env->iev_login == NULL)
+		fatal("calloc");
+
 	env->iev_fcgi = calloc(1, sizeof(*env->iev_fcgi));
 	if (env->iev_fcgi == NULL)
 		fatal("calloc");
 
+	env->iev_auth = calloc(env->prefork, sizeof(*env->iev_auth));
+	if (env->iev_auth == NULL)
+		fatal("calloc");
+
 	env->iev_gotweb = calloc(env->prefork, sizeof(*env->iev_gotweb));
 	if (env->iev_gotweb == NULL)
 		fatal("calloc");
 
 	spawn_process(env, argv0, env->iev_sockets,
-	    GOTWEBD_PROC_SOCKETS, gotwebd_username,
+	    GOTWEBD_PROC_SOCKETS, gotwebd_username, www_username,
 	    gotwebd_dispatch_server);
 
+	spawn_process(env, argv0, env->iev_login, GOTWEBD_PROC_LOGIN,
+	    gotwebd_username, www_username, gotwebd_dispatch_login);
+
 	spawn_process(env, argv0, env->iev_fcgi,
-	    GOTWEBD_PROC_FCGI, gotwebd_username,
+	    GOTWEBD_PROC_FCGI, gotwebd_username, www_username,
 	    gotwebd_dispatch_fcgi);
 
 	for (i = 0; i < env->prefork; ++i) {
+		spawn_process(env, argv0, &env->iev_auth[i],
+		    GOTWEBD_PROC_AUTH, gotwebd_username, www_username,
+		    gotwebd_dispatch_auth);
 		spawn_process(env, argv0, &env->iev_gotweb[i],
-		    GOTWEBD_PROC_GOTWEB, gotwebd_username,
+		    GOTWEBD_PROC_GOTWEB, gotwebd_username, www_username,
 		    gotwebd_dispatch_gotweb);
 	}
 
@@ -650,11 +867,17 @@ main(int argc, char **argv)
 static void
 connect_children(struct gotwebd *env)
 {
-	struct imsgev *iev_gotweb;
+	struct imsgev *iev_gotweb, *iev_auth;
 	int pipe[2];
 	int i;
+	int		 sock_flags = SOCK_STREAM | SOCK_NONBLOCK;
 
-	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe) == -1)
+#ifdef SOCK_CLOEXEC
+	sock_flags |= SOCK_CLOEXEC;
+#endif
+
+	if (socketpair(AF_UNIX, sock_flags,
+	    PF_UNSPEC, pipe) == -1)
 		fatal("socketpair");
 
 	if (main_compose_sockets(env, GOTWEBD_IMSG_CTL_PIPE, pipe[0], NULL, 0))
@@ -665,14 +888,30 @@ connect_children(struct gotwebd *env)
 
 	for (i = 0; i < env->prefork; i++) {
 		iev_gotweb = &env->iev_gotweb[i];
+		iev_auth = &env->iev_auth[i];
 
-		if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, pipe) == -1)
+		if (socketpair(AF_UNIX,
+		    SOCK_STREAM | sock_flags,
+		    PF_UNSPEC, pipe) == -1)
 			fatal("socketpair");
 
 		if (main_compose_sockets(env, GOTWEBD_IMSG_CTL_PIPE,
 		    pipe[0], NULL, 0))
 			fatal("send_imsg");
 
+		if (send_imsg(iev_auth, GOTWEBD_IMSG_CTL_PIPE,
+		    pipe[1], NULL, 0))
+			fatal("send_imsg");
+
+		if (socketpair(AF_UNIX,
+		    SOCK_STREAM | sock_flags,
+		    PF_UNSPEC, pipe) == -1)
+			fatal("socketpair");
+
+		if (send_imsg(iev_auth, GOTWEBD_IMSG_CTL_PIPE,
+		    pipe[0], NULL, 0))
+			fatal("send_imsg");
+
 		if (send_imsg(iev_gotweb, GOTWEBD_IMSG_CTL_PIPE,
 		    pipe[1], NULL, 0))
 			fatal("send_imsg");
@@ -684,18 +923,69 @@ gotwebd_configure(struct gotwebd *env, uid_t uid, gid_t gid)
 {
 	struct server *srv;
 	struct socket *sock;
+	struct gotwebd_repo *repo;
+	char secret[32];
+	int i;
 
 	/* gotweb need to reload its config. */
 	env->gotweb_pending = env->prefork;
+	env->auth_pending = env->prefork;
+
+	/* send global access rules */
+	for (i = 0; i < env->prefork; ++i) {
+		config_set_access_rules(&env->iev_auth[i],
+		    &env->access_rules);
+		config_set_access_rules(&env->iev_gotweb[i],
+		    &env->access_rules);
+	}
 
 	/* send our gotweb servers */
 	TAILQ_FOREACH(srv, &env->servers, entry) {
 		if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_SRV,
 		    -1, srv, sizeof(*srv)) == -1)
 			fatal("send_imsg GOTWEBD_IMSG_CFG_SRV");
+		if (main_compose_auth(env, GOTWEBD_IMSG_CFG_SRV,
+		    -1, srv, sizeof(*srv)) == -1)
+			fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SRV");
 		if (main_compose_gotweb(env, GOTWEBD_IMSG_CFG_SRV,
 		    -1, srv, sizeof(*srv)) == -1)
 			fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SRV");
+		if (main_compose_login(env, GOTWEBD_IMSG_CFG_SRV,
+		    -1, srv, sizeof(*srv)) == -1)
+			fatal("main_compose_gotweb GOTWEBD_IMSG_CFG_SRV");
+
+		/* send per-server access rules */
+		for (i = 0; i < env->prefork; ++i) {
+			config_set_access_rules(&env->iev_auth[i],
+			    &srv->access_rules);
+			config_set_access_rules(&env->iev_gotweb[i],
+			    &srv->access_rules);
+		}
+
+		/* send repositories and per-repository access rules */
+		TAILQ_FOREACH(repo, &srv->repos, entry) {
+			for (i = 0; i < env->prefork; i++) {
+				config_set_repository(&env->iev_auth[i],
+				    repo);
+				config_set_repository(&env->iev_gotweb[i],
+				    repo);
+
+				config_set_access_rules(&env->iev_auth[i],
+				    &repo->access_rules);
+				config_set_access_rules(&env->iev_gotweb[i],
+				    &repo->access_rules);
+			}
+		}
+
+		for (i = 0; i < env->prefork; i++) {
+			if (imsgbuf_flush(&env->iev_auth[i].ibuf) == -1)
+				fatal("imsgbuf_flush");
+			imsg_event_add(&env->iev_auth[i]);
+
+			if (imsgbuf_flush(&env->iev_gotweb[i].ibuf) == -1)
+				fatal("imsgbuf_flush");
+			imsg_event_add(&env->iev_gotweb[i]);
+		}
 	}
 
 	/* send our sockets */
@@ -711,6 +1001,50 @@ gotwebd_configure(struct gotwebd *env, uid_t uid, gid_t gid)
 	/* Connect servers and gotwebs. */
 	connect_children(env);
 
+	if (main_compose_auth(env, GOTWEBD_IMSG_AUTH_CONF, -1,
+	    &env->auth_config, sizeof(env->auth_config)) == -1)
+		fatal("send_imsg GOTWEB_IMSG_AUTH_CONF");
+	if (main_compose_gotweb(env, GOTWEBD_IMSG_AUTH_CONF, -1,
+	    &env->auth_config, sizeof(env->auth_config)) == -1)
+		fatal("main_compose_gotweb GOTWEB_IMSG_AUTH_CONF");
+
+	if (main_compose_auth(env, GOTWEBD_IMSG_WWW_UID, -1,
+	    &env->www_uid, sizeof(env->www_uid)) == -1)
+		fatal("main_compose_auth GOTWEB_IMSG_WWW_UID");
+	if (main_compose_gotweb(env, GOTWEBD_IMSG_WWW_UID, -1,
+	    &env->www_uid, sizeof(env->www_uid)) == -1)
+		fatal("main_compose_gotweb GOTWEB_IMSG_WWW_UID");
+
+	arc4random_buf(secret, sizeof(secret));
+
+	if (main_compose_login(env, GOTWEBD_IMSG_LOGIN_SECRET, -1,
+	    secret, sizeof(secret)) == -1)
+		fatal("main_compose_login GOTWEB_IMSG_LOGIN_SECRET");
+	if (main_compose_auth(env, GOTWEBD_IMSG_LOGIN_SECRET, -1,
+	    secret, sizeof(secret)) == -1)
+		fatal("main_compose_auth GOTWEB_IMSG_LOGIN_SECRET");
+
+	arc4random_buf(secret, sizeof(secret));
+
+	if (main_compose_auth(env, GOTWEBD_IMSG_AUTH_SECRET, -1,
+	    secret, sizeof(secret)) == -1)
+		fatal("main_compose_auth GOTWEB_IMSG_AUTH_SECRET");
+
+#ifdef __APPLE__
+	memset_s(secret, sizeof(*secret), 0, sizeof(*secret));
+#elif defined(__NetBSD__)
+	explicit_memset(secret, sizeof(*secret), 0);
+#else 
+	explicit_bzero(secret, sizeof(secret));
+#endif
+
+	if (login_privinit(env, uid, gid) == -1)
+		fatalx("cannot open authentication socket");
+
+	if (main_compose_login(env, GOTWEBD_IMSG_CFG_SOCK, env->login_sock->fd,
+	    NULL, 0) == -1)
+		fatal("main_compose_login GOTWEBD_IMSG_CFG_SOCK");
+
 	if (main_compose_sockets(env, GOTWEBD_IMSG_CFG_DONE, -1,
 	    NULL, 0) == -1)
 		fatal("send_imsg GOTWEBD_IMSG_CFG_DONE");
@@ -730,8 +1064,20 @@ gotwebd_configure_done(struct gotwebd *env)
 		if (env->gotweb_pending == 0 &&
 		    main_compose_gotweb(env, GOTWEBD_IMSG_CTL_START,
 		        -1, NULL, 0) == -1)
-			fatal("main_compose_sockets GOTWEBD_IMSG_CTL_START");
+			fatal("main_compose_gotweb GOTWEBD_IMSG_CTL_START");
 	}
+
+	if (env->auth_pending > 0) {
+		env->auth_pending--;
+		if (env->auth_pending == 0 &&
+		    main_compose_auth(env, GOTWEBD_IMSG_CTL_START,
+		        -1, NULL, 0) == -1)
+			fatal("main_compose_auth GOTWEBD_IMSG_CTL_START");
+	}
+
+	if (main_compose_login(env, GOTWEBD_IMSG_CTL_START,
+	    -1, NULL, 0) == -1)
+		fatal("send_imsg GOTWEBD_IMSG_CTL_START");
 }
 
 void
@@ -741,6 +1087,12 @@ gotwebd_shutdown(void)
 	pid_t		 pid;
 	int		 i, status;
 
+	event_del(&env->iev_login->ev);
+	imsgbuf_clear(&env->iev_login->ibuf);
+	close(env->iev_login->ibuf.fd);
+	env->iev_login->ibuf.fd = -1;
+	free(env->iev_login);
+
 	event_del(&env->iev_sockets->ev);
 	imsgbuf_clear(&env->iev_sockets->ibuf);
 	close(env->iev_sockets->ibuf.fd);
@@ -754,13 +1106,21 @@ gotwebd_shutdown(void)
 	free(env->iev_fcgi);
 
 	for (i = 0; i < env->prefork; ++i) {
+		event_del(&env->iev_auth[i].ev);
+		imsgbuf_clear(&env->iev_auth[i].ibuf);
+		close(env->iev_auth[i].ibuf.fd);
+		env->iev_auth[i].ibuf.fd = -1;
+
 		event_del(&env->iev_gotweb[i].ev);
 		imsgbuf_clear(&env->iev_gotweb[i].ibuf);
 		close(env->iev_gotweb[i].ibuf.fd);
 		env->iev_gotweb[i].ibuf.fd = -1;
 	}
+	free(env->iev_auth);
 	free(env->iev_gotweb);
 
+	free(env->login_sock);
+
 	do {
 		pid = waitpid(WAIT_ANY, &status, 0);
 		if (pid <= 0)
