@@ -39,6 +39,10 @@
 #define GOTWEBD_WWW_USER	 "www"
 #endif
 
+#define GOTWEBD_LOGIN_CMD	"weblogin"
+#define GOTWEBD_LOGIN_SOCKET	 "/var/run/gotweb-login.sock"
+#define GOTWEBD_LOGIN_TIMEOUT	 300 /* in seconds */
+
 #define GOTWEBD_MAXDESCRSZ	 1024
 #define GOTWEBD_MAXCLONEURLSZ	 1024
 #define GOTWEBD_CACHESIZE	 1024
@@ -55,6 +59,8 @@
 #define MAX_QUERYSTRING		 2048
 #define MAX_DOCUMENT_URI	 255
 #define MAX_SERVER_NAME		 255
+#define MAX_AUTH_COOKIE		 255
+#define MAX_IDENTIFIER_SIZE	 32
 
 #define GOTWEB_GIT_DIR		 ".git"
 
@@ -75,6 +81,7 @@
 #define D_SHOWDESC		 1
 #define D_SHOWURL		 1
 #define D_RESPECTEXPORTOK	 0
+#define D_HIDE_REPOSITORIES	 0
 #define D_MAXREPODISP		 25
 #define D_MAXSLCOMMDISP		 10
 #define D_MAXCOMMITDISP		 25
@@ -94,7 +101,6 @@
     (((n) + (FCGI_ALIGNMENT - 1)) & ~(FCGI_ALIGNMENT - 1))
 
 #define FD_RESERVE		 5
-#define FD_NEEDED		 6
 
 #define FCGI_BEGIN_REQUEST	 1
 #define FCGI_ABORT_REQUEST	 2
@@ -125,6 +131,8 @@ enum gotwebd_proc_type {
 	GOTWEBD_PROC_PARENT,
 	GOTWEBD_PROC_SOCKETS,
 	GOTWEBD_PROC_FCGI,
+	GOTWEBD_PROC_LOGIN,
+	GOTWEBD_PROC_AUTH,
 	GOTWEBD_PROC_GOTWEB,
 };
 
@@ -132,11 +140,17 @@ enum imsg_type {
 	GOTWEBD_IMSG_CFG_SRV,
 	GOTWEBD_IMSG_CFG_SOCK,
 	GOTWEBD_IMSG_CFG_FD,
+	GOTWEBD_IMSG_CFG_ACCESS_RULE,
+	GOTWEBD_IMSG_CFG_REPO,
 	GOTWEBD_IMSG_CFG_DONE,
 	GOTWEBD_IMSG_CTL_PIPE,
 	GOTWEBD_IMSG_CTL_START,
+	GOTWEBD_IMSG_LOGIN_SECRET,
+	GOTWEBD_IMSG_AUTH_SECRET,
+	GOTWEBD_IMSG_AUTH_CONF,
 	GOTWEBD_IMSG_FCGI_PARSE_PARAMS,
 	GOTWEBD_IMSG_FCGI_PARAMS,
+	GOTWEBD_IMSG_WWW_UID,
 	GOTWEBD_IMSG_REQ_ABORT,
 	GOTWEBD_IMSG_REQ_PROCESS,
 };
@@ -275,6 +289,7 @@ struct querystring {
 	char		 headref[MAX_DOCUMENT_URI];
 	int		 index_page;
 	char		 path[PATH_MAX];
+	char		 login[MAX_AUTH_COOKIE];
 };
 
 struct gotwebd_fcgi_params {
@@ -282,6 +297,7 @@ struct gotwebd_fcgi_params {
 	struct querystring		 qs;
 	char				 document_uri[MAX_DOCUMENT_URI];
 	char				 server_name[MAX_SERVER_NAME];
+	char				 auth_cookie[MAX_AUTH_COOKIE];
 	int				 https;
 };
 
@@ -312,6 +328,9 @@ struct request {
 	int				 nparams_parsed;
 
 	int				 client_status;
+
+	uid_t				 client_uid;
+	char				 access_identifier[MAX_IDENTIFIER_SIZE];
 };
 TAILQ_HEAD(requestlist, request);
 
@@ -339,6 +358,38 @@ struct address {
 };
 TAILQ_HEAD(addresslist, address);
 
+enum gotwebd_auth_config {
+	GOTWEBD_AUTH_DISABLED	= 0xf00000ff,
+	GOTWEBD_AUTH_SECURE	= 0x00808000,
+	GOTWEBD_AUTH_INSECURE	= 0x0f7f7f00
+};
+
+enum gotwebd_access {
+	GOTWEBD_ACCESS_NO_MATCH = -2,
+	GOTWEBD_ACCESS_DENIED = -1,
+	GOTWEBD_ACCESS_PERMITTED = 1
+};
+
+struct gotwebd_access_rule {
+	STAILQ_ENTRY(gotwebd_access_rule) entry;
+
+	enum gotwebd_access access;
+	char identifier[MAX_IDENTIFIER_SIZE];
+};
+STAILQ_HEAD(gotwebd_access_rule_list, gotwebd_access_rule);
+
+struct gotwebd_repo {
+	TAILQ_ENTRY(gotwebd_repo)	 entry;
+
+	char name[NAME_MAX];
+
+	enum gotwebd_auth_config	auth_config;
+	struct gotwebd_access_rule_list access_rules;
+
+	int				hidden;
+};
+TAILQ_HEAD(gotwebd_repolist, gotwebd_repo);
+
 struct server {
 	TAILQ_ENTRY(server)	 entry;
 
@@ -351,6 +402,7 @@ struct server {
 	char		 logo[GOTWEBD_MAXTEXT];
 	char		 logo_url[GOTWEBD_MAXTEXT];
 	char		 custom_css[PATH_MAX];
+	char		 login_hint_user[MAX_IDENTIFIER_SIZE];
 
 	size_t		 max_repos_display;
 	size_t		 max_commits_display;
@@ -363,6 +415,12 @@ struct server {
 	int		 show_repo_description;
 	int		 show_repo_cloneurl;
 	int		 respect_exportok;
+	int		 hide_repositories;
+
+	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list access_rules;
+
+	struct gotwebd_repolist	 repos;
 };
 TAILQ_HEAD(serverlist, server);
 
@@ -401,6 +459,12 @@ struct gotwebd {
 	struct socketlist	sockets;
 	struct addresslist	addresses;
 
+	struct socket	*login_sock;
+	struct event	 login_pause_ev;
+
+	enum gotwebd_auth_config auth_config;
+	struct gotwebd_access_rule_list access_rules;
+
 	int		 pack_fds[GOTWEB_PACK_NUM_TEMPFILES];
 	int		 priv_fd[PRIV_FDS__MAX];
 
@@ -414,13 +478,20 @@ struct gotwebd {
 	struct imsgev	*iev_parent;
 	struct imsgev	*iev_sockets;
 	struct imsgev	*iev_fcgi;
+	struct imsgev	*iev_login;
+	struct imsgev	*iev_gotsh;
+	struct imsgev	*iev_auth;
 	struct imsgev	*iev_gotweb;
 
 	uint16_t	 prefork;
+	int		 auth_pending;
 	int		 gotweb_pending;
 	int		 *worker_load;
 
 	char		 httpd_chroot[PATH_MAX];
+	uid_t		 www_uid;
+
+	char		 login_hint_user[MAX_IDENTIFIER_SIZE];
 };
 
 /*
@@ -455,6 +526,7 @@ enum querystring_elements {
 	HEADREF,
 	INDEX_PAGE,
 	PATH,
+	LOGIN,
 };
 
 extern struct gotwebd	*gotwebd_env;
@@ -468,8 +540,12 @@ int	 imsg_compose_event(struct imsgev *, uint16_t, uint32_t,
 	    pid_t, int, const void *, size_t);
 int	 main_compose_sockets(struct gotwebd *, uint32_t, int,
 	    const void *, uint16_t);
+int	 main_compose_login(struct gotwebd *, uint32_t, int,
+	    const void *, uint16_t);
 int	 sockets_compose_main(struct gotwebd *, uint32_t,
 	    const void *, uint16_t);
+int	 main_compose_auth(struct gotwebd *, uint32_t, int,
+	    const void *, uint16_t);
 int	 main_compose_gotweb(struct gotwebd *, uint32_t, int,
 	    const void *, uint16_t);
 
@@ -477,10 +553,26 @@ int	 main_compose_gotweb(struct gotwebd *, uint32_t, int,
 void sockets(struct gotwebd *, int);
 void sockets_parse_sockets(struct gotwebd *);
 void sockets_socket_accept(int, short, void *);
+struct socket *sockets_conf_new_socket(int, struct address *);
 int sockets_privinit(struct gotwebd *, struct socket *, uid_t, gid_t);
 void sockets_rlimit(int);
 
+/* login.c */
+char *login_gen_token(uint64_t, const char *, time_t, const char *, size_t,
+    const char *);
+int login_check_token(uid_t *, char **, const char *, const char *, size_t,
+    const char *);
+int login_privinit(struct gotwebd *, uid_t, gid_t);
+void gotwebd_login(struct gotwebd *, int);
+
+/* auth.c */
+void gotwebd_auth(struct gotwebd *, int);
+
 /* gotweb.c */
+struct server *gotweb_get_server(const char *);
+struct gotwebd_repo * gotweb_get_repository(struct server *, const char *);
+int gotweb_reply(struct request *c, int status, const char *ctype,
+    struct gotweb_url *);
 void gotweb_index_navs(struct request *, struct gotweb_url *, int *,
     struct gotweb_url *, int *);
 int gotweb_render_age(struct template *, time_t);
@@ -512,8 +604,10 @@ int	gotweb_render_summary(struct template *);
 int	gotweb_render_blame(struct template *);
 int	gotweb_render_patch(struct template *);
 int	gotweb_render_rss(struct template *);
+int	gotweb_render_unauthorized(struct template *);
 
 /* parse.y */
+struct gotwebd_repo * gotwebd_new_repo(const char *);
 int parse_config(const char *, struct gotwebd *);
 int cmdline_symset(char *);
 
@@ -550,4 +644,11 @@ int config_getsock(struct gotwebd *, struct imsg *);
 int config_setfd(struct gotwebd *);
 int config_getfd(struct gotwebd *, struct imsg *);
 int config_getcfg(struct gotwebd *, struct imsg *);
+void config_set_access_rules(struct imsgev *,
+    struct gotwebd_access_rule_list *);
+void config_get_access_rule(struct gotwebd_access_rule_list *, struct imsg *);
+void config_free_access_rules(struct gotwebd_access_rule_list *);
+void config_set_repository(struct imsgev *, struct gotwebd_repo *);
+void config_get_repository(struct gotwebd_repolist *, struct imsg *);
+void config_free_repos(struct gotwebd_repolist *);
 int config_init(struct gotwebd *);
