@@ -8,7 +8,7 @@
 #
 # Intended to be called from configure.ac (via autogen.sh)
 GOT_RELEASE=yes
-GOT_PORTABLE_VER=0.118.1
+GOT_PORTABLE_VER=0.120
 
 [ -d ".git" -a "$GOT_RELEASE" = "no" ] || { echo "$GOT_PORTABLE_VER" ; exit ; }
 
